//------------------------------------------------------------------------------
//
//   Greg's Blogging Mod - XML-RPC protocol
//
//------------------------------------------------------------------------------

package gcewing.blogging.web;

import java.util.*;
import org.w3c.dom.*;
import sun.misc.BASE64Encoder;

import gcewing.blogging.*;

public class XMLRPC {

	static boolean debugCallText = true;
	static boolean debugResponseText = true;
	static boolean debugResponseData = true;

	public static Array callMethod(String url, String name, Object... args) {
		XML call = new MethodCall(name, args);
		if (debugCallText) {
			System.out.printf("------- Begin XML-RPC Call Text -------\n");
			System.out.print(call.toString());
			System.out.printf("-------- End XML-RPC Call Text --------\n");
		}
		HTTP http = new HTTP(url, "POST", call.toString());
		if (http.code != 200)
			throw new EnvironmentError(http.message);
		XML xml = XML.parse(http.body.replace("&amp;", "&"));
		if (debugResponseText) {
			System.out.printf("------- Begin XML-RPC Response Text -------\n");
			System.out.print(xml.toString());
			System.out.printf("-------- End XML-RPC Response Text --------\n");
		}
		Response response = parseResponse(xml);
		if (debugResponseData)
			System.out.printf("XML-RPC Response: %s\n", response);
		if (response instanceof MethodResponse)
			return ((MethodResponse)response).params;
		else {
			FaultResponse fault = (FaultResponse)response;
			throw new EnvironmentError(
				String.format("XML-RPC Fault: %d: %s", fault.faultCode, fault.faultString));
		}
	}

	static class MethodCall extends XML {
	
		public MethodCall(String name, Object... args) {
			super("methodCall");
			add(root, elem("methodName", name), params(args));
		}
		
		Element params(Object[] args) {
			Element node = elem("params");
			for (Object arg : args) {
				Element child = elem("param", elem("value", value(arg)));
				node.appendChild(child);
			}
			return node;
		}
		
		Element value(Object x) {
			String type;
			if (x instanceof Array)
				return array((Array)x);
			if (x instanceof Struct)
				return struct((Struct)x);
			if (x instanceof byte[])
				return elem("base64", new BASE64Encoder().encode((byte[])x));
			if (x instanceof String)
				type = "string";
			else if (x instanceof Integer)
				type = "int";
			else if (x instanceof Boolean)
				type = "boolean";
			else
				throw new EnvironmentError(
					String.format("Unsupported XML-RPC parameter type %s\n", x.getClass().getName()));
			return elem(type, x.toString());
		}
		
		Element array(Array a) {
			Element data = elem("data");
			for (Object x : a)
				add(data, value(x));
			return elem("array", data);
		}
		
		Element struct(Struct s) {
			Element result = elem("struct");
			for (String name : s.keySet())
				add(result, elem("member", elem("name", name), value(s.get(name))));
			return result;
		}
	
	}
	
	public static class Response {
	}
	
	public static class MethodResponse extends Response {
		Array params;
		
		public String toString() {
			return String.format("MethodResponse(%s)", this.params.toString("", ""));
		}
		
	}
	
	public static class FaultResponse extends Response {
		int faultCode;
		String faultString;
		
		public String toString() {
			return String.format("FaultResponse(faultCode=%s,faultString=%s)",
				faultCode, asString(faultString));
		}

	}
	
	@SuppressWarnings("unchecked")
	static <T> T asType(Object x, Class<T> c) {
		if (c.isAssignableFrom(x.getClass()))
			return (T)x;
		else
			throw new EnvironmentError(
				String.format("Type error in XML-RPC response, expected %s, got %s",
					c.getSimpleName(), x.getClass().getSimpleName()));
	}
	
	public static String asString(Object value) {
		if (value instanceof String)
			return "\"" + value.toString() + "\"";
		else
			return value.toString();
	}
	
	static Response parseResponse(XML xml) {
		expect(xml.root, "methodResponse");
		Node child = firstElement(xml.root);
		String type = child.getNodeName();
		if (type.equals("params"))
			return parseMethodResponse(child);
		if (type.equals("fault"))
			return parseFault(child);
		throw new EnvironmentError(String.format("Unknown XML-RPC response element <%s>", type));
	}
	
	static MethodResponse parseMethodResponse(Node paramsNode) {
		MethodResponse result = new MethodResponse();
		result.params = new Array();
		for (Node p = paramsNode.getFirstChild(); p != null; p = p.getNextSibling())
			if (isElement(p))
				result.params.add(parseParam(p));
		return result;
	}
	
	static FaultResponse parseFault(Node faultNode) {
		Node valueNode = find(faultNode, "value");
		Object value = parseValue(valueNode);
		FaultResponse result = new FaultResponse();
		try {
			Struct map = (Struct)value;
			result.faultCode = (Integer)map.get("faultCode");
			result.faultString = (String)map.get("faultString");
		}
		catch (ClassCastException e) {
			throw new EnvironmentError("Malformed XML-RPC fault response");
		}
		return result;
	}
	
	static Object parseParam(Node paramNode) {
		expect(paramNode, "param");
		Node valueNode = firstElement(paramNode);
		return parseValue(valueNode);
	}
	
	static Object parseValue(Node valueNode) {
		expect(valueNode, "value");
		Node typeNode = firstElement(valueNode);
		return parseType(typeNode);
	}
	
	static Object parseType(Node typeNode) {
		String type = typeNode.getNodeName();
		String data = typeNode.getTextContent();
		try {
			//if (type.equals("string"))
			//	return data;
			if (type.startsWith("i"))
				return Integer.parseInt(data);
			if (type.equals("boolean"))
				return Integer.parseInt(data) != 0;
			if (type.equals("array"))
				return parseArray(typeNode);
			if (type.equals("struct"))
				return parseStruct(typeNode);
			return data;
		}
		catch (NumberFormatException e) {
			throw new EnvironmentError(String.format("Invalid <%s> value in XML-RPC result: '%s'",
				type, data));
		}
		catch (Exception e) {
			throw new EnvironmentError(e);
		}
		//throw new EnvironmentError(String.format("Unknown type in XML-RPC result: <%s>", type));
	}
	
	static Array parseArray(Node arrayNode) {
		Node dataNode = arrayNode.getFirstChild();
		expect(dataNode, "data");
		Array result = new Array();
		for (Node v = dataNode.getFirstChild(); v != null; v = v.getNextSibling())
			if (isElement(v))
				result.add(parseValue(v));
		return result;
	}
	
	static Struct parseStruct(Node structNode) {
		Struct result = new Struct();
		for (Node m = structNode.getFirstChild(); m != null; m = m.getNextSibling())
			if (isElement(m))
				parseMember(m, result);
		return result;
	}
	
	static void parseMember(Node memberNode, Struct map) {
		expect(memberNode, "member");
		Node nameNode = find(memberNode, "name");
		Node valueNode = find(memberNode, "value");
		String name = parseName(nameNode);
		Object value = parseValue(valueNode);
		map.put(name, value);
	}
	
	static String parseName(Node nameNode) {
		expect(nameNode, "name");
		return nameNode.getTextContent();
	}
	
	static void expect(Node node, String name) {
		if (node.getNodeName() != name)
			throw new EnvironmentError(
				String.format("Invalid element <%s> in XML-RPC response, expected <%s>",
					node.getNodeName(), name));
	}
	
	static Node find(Node parent, String name) {
		for (Node n = parent.getFirstChild(); n != null; n = n.getNextSibling())
			if (n.getNodeName().equals(name))
				return n;
		throw new EnvironmentError(
			String.format("Expected element <%s> not found in element <%s> of XML-RPC response",
				name, parent.getNodeName()));
	}
	
	static Node firstElement(Node parent) {
		for (Node n = parent.getFirstChild(); n != null; n = n.getNextSibling())
			if (isElement(n))
				return n;
		throw new EnvironmentError(
			String.format("Element <%s> of XML-RPC response has no child element",
				parent.getNodeName()));
	}
	
	static boolean isElement(Node node) {
		//return node.getNodeType() == Node.ELEMENT_NODE;
		return node.getNodeName() != "#text";
	}

}
